from astropy.io import ascii
import numpy as np
import sys

#Copyright 2015 Emiliano Merlin
#
#This file is part of T-PHOT.
#
#T-PHOT is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#T-PHOT is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with T-PHOT.  If not, see <http://www.gnu.org/licenses/>.

"""
Script to make output LRI catalog the same length as any input catalog
adding fake entries for missing sources
"""

def adapt(HRIcat,LRIcat,OUTcat):

    fff=open(LRIcat,'r')
    lines=fff.readlines()
    out=open(OUTcat,'w')
    out.write("#  1   ObjectID\n#  2   X\n#  3   Y\n#  4   Cell \n#  5   cx          X pos in cell \n#  6   cy          Y pos in cell \n#  7   Rcell \n#  8   FitQty      Fit Quantity (Flux or Scale Factor)\n#  9   FitQuErr    Variance of Fit Quantity\n# 10   SExF        SExtractor flux_best in hires catalog\n# 11   Totflux     Total flux in hires cutout\n# 12   Flag       Flag for saturation and/or prior blending\n# 13   NumFits     Number of fits for this object\n# 14   MaxFlag     Max flag value in object template\n#  15  MaxCvID\n#  16  MaxCvRatio\n")

    hrcat=ascii.read(HRIcat)
    lrcat=ascii.read(LRIcat)
    columns=np.size(lrcat.colnames)
    IDsLR=list(lrcat[lrcat.colnames[0]])

    for i in range(np.size(hrcat)):
        id=hrcat[i][0]
        x=hrcat[i][1]
        y=hrcat[i][2]
        if id in IDsLR:
            idx=IDsLR.index(id)
            out.write(lines[idx+16])
        else:
            out.write("%d %f %f 0 0.0000 0.0000 0.0000 -0.990000E+02 0.990000E+10 %e %e 100 0 -99 0 99.0000\n"%(id,x,y,hrcat[i][8],hrcat[i][8]))

    out.close()

    return(0)

if __name__ == '__main__':

    if np.size(sys.argv)!=4:
        print "USAGE: python adapt_cats.py INPUTcat LRIcat OUTcat"
        sys.exit()

    adapt(sys.argv[1],sys.argv[2],sys.argv[3])

    
